/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Connect to PostgreSQL
 *
 * Copyright (C) 2004, Constantine Filin and Christos Ricudis
 *
 * Christos Ricudis <ricudis@itc.auth.gr>
 * Constantine Filin <cf@intermedia.net>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/linkedlists.h>
#include <asterisk/chanvars.h>
#include <asterisk/lock.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>

#include <astodbc/astodbc.h>

static char *tdesc = "Simple ODBC Interface";

static char *app = "ODBC";

static char *synopsis = "Do several ODBC things";

static char *descrip =
"ODBC():  Do several ODBC things\n"
"Syntax:\n"
"  ODBC(Connect connid db_conn_string [poolable [writeable]])\n"
"    Connects to a database.  db_conn_string should look like\n"
"    Driver=dbDriver;Server=dbServer;Port=dbPort;Database=dbName;UID=dbUser;PWD=dbPassword\n"
"    It is passed to SQLDriverConnect(). Connection identifer is returned in ${connid}\n"
"  ODBC(Query resultid ${connid} query_string)\n"
"    Executes standard ODBC query contained in query-string using established\n"
"    connection identified by ${connid}. Result of query is stored in ${resultid}.\n"
"  ODBC(Fetch fetchid ${resultid} var1 var2 ... varN)\n"
"    Fetches a single row from a result set contained in ${resultid}.\n"
"    Assigns returned fields to ${var1} ... ${varn}.  ${fetchid} is set TRUE\n"
"    if additional rows exist in result set.\n"
"  ODBC(Clear ${resultid})\n"
"    Frees memory and data structures associated with result set.\n"
"  ODBC(Disconnect ${connid})\n"
"    Disconnects from named connection to ODBC.\n" ;

/*
EXAMPLES OF USE :

exten => s,2,ODBC(Connect connid Driver=PostrgreSQL;Server=dbServer;Port=5432;Database=dbName;UID=dbUser;PWD=dbPassword)
exten => s,3,ODBC(Query resultid ${connid} SELECT username,credit FROM credit WHERE callerid=${CALLERIDNUM})
exten => s,4,ODBC(Fetch fetchid ${resultid} datavar1 datavar2)
exten => s,5,GotoIf(${fetchid}?6:8)
exten => s,6,Festival("User ${datavar1} currently has credit balance of ${datavar2} dollars.")
exten => s,7,Goto(s,4)
exten => s,8,ODBC(Clear ${resultid})
exten => s,9,ODBC(Disconnect ${connid})
*/

STANDARD_LOCAL_USER;
LOCAL_USER_DECL;

extern void pbx_builtin_setvar_helper(struct ast_channel *chan, char *name, char *value);

#define AST_ODBC_ID_DUMMY   0
#define AST_ODBC_ID_CONNID  1
#define AST_ODBC_ID_RESID   2
#define AST_ODBC_ID_FETCHID 3

struct ast_ODBC_id {
        int             identifier_type; /* 0=dummy, 1=connid, 2=resultid */
        int             identifier;
        void*           data;
        AST_LIST_ENTRY(ast_ODBC_id) entries;
} *ast_ODBC_id;

AST_LIST_HEAD(ODBCidshead,ast_ODBC_id) _odbc_ids_head;

/* helpful procs */
static void *find_identifier(int identifier,int identifier_type) {
        struct ODBCidshead *headp;
        struct ast_ODBC_id *i;
	void *res=NULL;
	int found=0;

        headp=&_odbc_ids_head;

	if (AST_LIST_LOCK(headp)) {
		ast_log(LOG_WARNING,"Unable to lock identifiers list\n");
	} else {
		AST_LIST_TRAVERSE(headp,i,entries) {
			if ((i->identifier==identifier) && (i->identifier_type==identifier_type)) {
				found=1;
				res=i->data;
				break;
			}
		}
		if (!found) {
			ast_log(LOG_WARNING,"Identifier %d, identifier_type %d not found in identifier list\n",identifier,identifier_type);
		}
		AST_LIST_UNLOCK(headp);
	}

	return res;
}

static int add_identifier(int identifier_type,void *data) {
        struct ast_ODBC_id *i,*j;
        struct ODBCidshead *headp;
	int maxidentifier=0;

        headp=&_odbc_ids_head;
	i=NULL;
	j=NULL;

	if (AST_LIST_LOCK(headp)) {
		ast_log(LOG_WARNING,"Unable to lock identifiers list\n");
		return(-1);
	} else {
                i=malloc(sizeof(struct ast_ODBC_id));
		AST_LIST_TRAVERSE(headp,j,entries) {
			if (j->identifier>maxidentifier) {
				maxidentifier=j->identifier;
			}
		}
		i->identifier=maxidentifier+1;
		i->identifier_type=identifier_type;
		i->data=data;
		AST_LIST_INSERT_HEAD(headp,i,entries);
		AST_LIST_UNLOCK(headp);
	}
	return i->identifier;
}

static int del_identifier(int identifier,int identifier_type) {
        struct ast_ODBC_id *i;
        struct ODBCidshead *headp;
	int found=0;

        headp=&_odbc_ids_head;

        if (AST_LIST_LOCK(headp)) {
		ast_log(LOG_WARNING,"Unable to lock identifiers list\n");
	} else {
		AST_LIST_TRAVERSE(headp,i,entries) {
			if ((i->identifier==identifier) &&
			    (i->identifier_type==identifier_type)) {
                                AST_LIST_REMOVE(headp,i,ast_ODBC_id,entries);
				free(i);
				found=1;
				break;
			}
		}
		AST_LIST_UNLOCK(headp);
	}

	if (found==0) {
		ast_log(LOG_WARNING,"Could not find identifier %d, identifier_type %d in list to delete\n",identifier,identifier_type);
		return(-1);
	} else {
		return(0);
	}
}

static int set_asterisk_int(struct ast_channel *chan, char *varname, int id) {
	if( id>=0 ) {
		char s[100] = "";
		snprintf(s, sizeof(s)-1, "%d", id);
#if _DEBUG
                ast_log(LOG_WARNING,"ODBC: setting var '%s' to value '%s'\n",varname,s);
#endif
		pbx_builtin_setvar_helper(chan,varname,s);
	}
	return id;
}

static int add_identifier_and_set_asterisk_int(struct ast_channel *chan, char *varname, int identifier_type, void *data) {
	return set_asterisk_int(chan,varname,add_identifier(identifier_type,data));
}

static char* strsep_with_many_delims( char** stringp, const char* delim ) {
        char* result = strsep(stringp,delim);
        if( stringp && *stringp ) {
                // keep moving it until it is not a delimiter
                while( strchr(delim,**stringp) ) {
                        stringp[0]++;
                }
        }
        return result;
}

static int safe_scan_int( char** data, char* delim, int def ) {
	char* end;
	int res = def;
        char* s = strsep_with_many_delims(data,delim);
	if( s ) {
		res = strtol(s,&end,10);
		if (*end) res = def;  /* not an integer */
	}
	return res;
}

/* ODBC operations */
static int aODBC_connect(struct ast_channel *chan, char *data) {

        const char*     astodbc_result;
        void*           astodbc_conn;

        char*           connid_var;
        char*           db_conn_string;
        int             poolable;
        int             writeable;

        strsep_with_many_delims(&data," "); // eat the first token, we already know it :P

        connid_var      = strsep_with_many_delims(&data," ");
        db_conn_string  = strsep_with_many_delims(&data," \n");
        poolable        = safe_scan_int(&data," \n",1);
        writeable       = safe_scan_int(&data," \n",1);

        if( connid_var && db_conn_string ) {
                astodbc_result = ast_odbc_conn_new(db_conn_string,poolable,writeable,&astodbc_conn);
                if( strlen(astodbc_result)==0 ) {
                        add_identifier_and_set_asterisk_int(chan,connid_var,AST_ODBC_ID_CONNID,astodbc_conn);
                        return 0;
		}
		else {
                        ast_log(LOG_WARNING,"aODBC_connect: ast_odbc_conn_new(%s) returned '%s'\n",db_conn_string,astodbc_result);
		}
 	}
	else {
                ast_log(LOG_WARNING,"aODBC_connect: missing some arguments\n");
	}

	return -1;
}

static int aODBC_query(struct ast_channel *chan, char *data) {

        const char*     astodbc_result;
        void*           astodbc_conn;
        void*           astodbc_query;

        char*           resultid_var;
        int             connid;
        char*           querystring;

        strsep_with_many_delims(&data," "); // eat the first token, we already know it :P

        resultid_var = strsep_with_many_delims(&data," ");
	connid       = safe_scan_int(&data," ",-1);
        querystring  = data;

        if( resultid_var && (connid>=0) && querystring ) {
                if( (astodbc_conn=find_identifier(connid,AST_ODBC_ID_CONNID))!=NULL ) {
                        astodbc_result = ast_odbc_query_new(astodbc_conn,querystring,&astodbc_query);
                        if( strlen(astodbc_result)==0 ) {
                                add_identifier_and_set_asterisk_int(chan,resultid_var,AST_ODBC_ID_RESID,astodbc_query);
				return 0;
			}
			else {
                                ast_log(LOG_WARNING,"aODBC_query: ast_odbc_query_new(%s) returned '%s'\n",querystring,astodbc_result);
			}
		}
		else {
                        ast_log(LOG_WARNING,"aODBC_query: Invalid connection identifier %d passed in aODBC_query\n",connid);
		}
	}
	else {
                ast_log(LOG_WARNING,"aODBC_query: missing some arguments\n");
	}

	return -1;
}

static int aODBC_fetch(struct ast_channel *chan, char *data) {

        const char*     astodbc_result;
        void*           astodbc_query;

        char*           fetchid_var;
        int             resultid;

        strsep_with_many_delims(&data," "); // eat the first token, we already know it :P

        fetchid_var = strsep_with_many_delims(&data," ");
	resultid    = safe_scan_int(&data," ",-1);

        if( fetchid_var && (resultid>=0) ) {
                if( (astodbc_query=find_identifier(resultid,AST_ODBC_ID_RESID))!=NULL ) {
                        astodbc_result = ast_odbc_query_fetch(astodbc_query);
                        if( strlen(astodbc_result)==0 ) {
                                if( !ast_odbc_query_eof(astodbc_query) ) {
                                        int numFields,j;
                                        numFields = ast_odbc_query_get_column_count(astodbc_query);
                                        for (j=0;j<numFields;j++) {
                                                char* varname = strsep_with_many_delims(&data," ");
                                                if( varname==NULL ) {
                                                        ast_log(LOG_WARNING,"ast_ODBC_fetch: More fields (%d) than variables (%d)\n",numFields,j);
                                                        break;
                                                }
                                                pbx_builtin_setvar_helper(chan,varname,(char*)ast_odbc_query_getstring_by_ndx(astodbc_query,j,"NULL"));
                                        }
#ifdef _DEBUG
                                        ast_log(LOG_WARNING,"ast_ODBC_fetch: numFields=%d\n",numFields);
#endif
                                        set_asterisk_int(chan,fetchid_var,1); // try more rows
                                } else {
#if _DEBUG
                                        ast_log(LOG_WARNING,"ast_ODBC_fetch : EOF\n");
#endif
                                        set_asterisk_int(chan,fetchid_var,0); // no more rows
                                }
                                return 0;
                        }
                        else {
                                ast_log(LOG_WARNING,"aODBC_fetch: ast_odbc_query_fetch() returned %s\n",astodbc_result);
                        }
		}
		else {
                        ast_log(LOG_WARNING,"aODBC_fetch: Invalid result identifier %d passed\n",resultid);
		}
	}
	else {
                ast_log(LOG_WARNING,"aODBC_fetch: missing some arguments\n");
	}

	return -1;
}

static int aODBC_clear(struct ast_channel *chan, char *data) {

        void* astodbc_query;

	int id;

        strsep_with_many_delims(&data," "); // eat the first token, we already know it :P

	id = safe_scan_int(&data," \n",-1);

        if( (astodbc_query=find_identifier(id,AST_ODBC_ID_RESID))!=NULL ) {
                astodbc_query = ast_odbc_query_delete(astodbc_query);
                del_identifier(id,AST_ODBC_ID_RESID);
	} else {
                ast_log(LOG_WARNING,"Invalid result identifier %d passed in aODBC_clear\n",id);
	}

	return 0;
}

static int aODBC_disconnect(struct ast_channel *chan, char *data) {

        void* astodbc_conn;

	int id;

        strsep_with_many_delims(&data," "); // eat the first token, we already know it :P

	id = safe_scan_int(&data," \n",-1);

        if( (astodbc_conn=find_identifier(id,AST_ODBC_ID_CONNID))!=NULL ) {
                astodbc_conn = ast_odbc_conn_delete(astodbc_conn);
                del_identifier(id,AST_ODBC_ID_CONNID);
	} else {
                ast_log(LOG_WARNING,"Invalid connection identifier %d passed in aODBC_disconnect\n",id);
	}

	return 0;
}

static int ODBC_exec(struct ast_channel *chan, void *data) {
	struct localuser *u;
	int result;

#if _DEBUG
        fprintf(stderr,"ODBC_exec: data=%s\n",(char*)data);
#endif

	if (!data) {
		ast_log(LOG_WARNING, "APP_MYSQL requires an argument (see manual)\n");
		return -1;
	}

	LOCAL_USER_ADD(u);
	result=0;

	if (strncasecmp("connect",data,strlen("connect"))==0) {
                result = aODBC_connect(chan,ast_strdupa(data));
	} else 	if (strncasecmp("query",data,strlen("query"))==0) {
                result = aODBC_query(chan,ast_strdupa(data));
	} else 	if (strncasecmp("fetch",data,strlen("fetch"))==0) {
                result = aODBC_fetch(chan,ast_strdupa(data));
	} else 	if (strncasecmp("clear",data,strlen("clear"))==0) {
                result = aODBC_clear(chan,ast_strdupa(data));
	} else 	if (strncasecmp("disconnect",data,strlen("disconnect"))==0) {
                result = aODBC_disconnect(chan,ast_strdupa(data));
	} else {
                ast_log(LOG_WARNING, "Unknown argument to ODBC application : %s\n",(char *)data);
		result=-1;
	}

	LOCAL_USER_REMOVE(u);
	return result;

}

int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
	return ast_unregister_application(app);
}

int load_module(void)
{
        struct ODBCidshead *headp = &_odbc_ids_head;
	AST_LIST_HEAD_INIT(headp);
        return ast_register_application(app,ODBC_exec,synopsis,descrip);
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
